local ADDON_NAME="AVR"
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, true)

local _G = _G
local pairs = pairs
local ipairs = ipairs
local remove = table.remove
local insert = table.insert
local max = math.max
local min = math.min
local abs = math.abs
local atan2 = math.atan2
local sin = math.sin
local cos = math.cos
local type = type
local select = select
local next = next
local GetTime = GetTime
local deg2rad = math.pi/180.0
local pi = math.pi
local tostring,tonumber=tostring,tonumber
local strlen=string.len

local Core=AVR

local raidID={}
local partyID={}
for i=1,40 do raidID[i]="raid"..i end
for i=1,4 do partyID[i]="party"..i end


AVRRangeWarningMesh={Embed=Core.Embed}
function AVRRangeWarningMesh:New(range,radius)
	if self ~= AVRRangeWarningMesh then return end
	local s=AVRMesh:New()
	AVRRangeWarningMesh:Embed(s)
	s.range=range or 10
	s.radius=radius or 1
	s.rangeSq=s.range*s.range
	s.numMarkers=0
	s.spokes=6
	return s
end

function AVRRangeWarningMesh:SetRange(range)
	self.range=range
	self.rangeSq=self.range*self.range
	self.vertices=nil
	return self
end

function AVRRangeWarningMesh:SetRadius(radius)
	self.radius=radius
	self.vertices=nil
	return self
end

function AVRRangeWarningMesh:GetOptions()
	local o=AVRMesh.GetOptions(self)
	o.args.rangeWarning = {
		type = "group",
		name = L["Range warning properties"],
		inline = true,
		order = 80,
		args = {
			range = {
				type = "range",
				name = L["Range"],
				order = 10,
				width = "full",
				min = 0.0, max=500, bigStep=1.0
			},
			radius = {
				type = "range",
				name = L["Radius"],
				order = 10,
				width = "full",
				min = 0.0, max=10, bigStep=0.5
			}
		}
	}
	return o
end

function AVRRangeWarningMesh:GenerateMarkerMesh()
	local a,x,y,v,t
	
	local origin=self:AddVertex(0,0,0)
	local o=self:AddVertex(self.radius,0,0)
	
	for i=1,self.spokes do
		a=pi*2/self.spokes*i
		
		x=cos(a)*self.radius
		y=sin(a)*self.radius
		v=self:AddVertex(x,y,0)
		
		
		if (i%2)==0 then
			t=AVRTriangle:New(origin,o,v,nil,0,0,0)
		else
			t=AVRTriangle:New(origin,o,v)
		end
		self:AddTriangle(t)
		
		o=v
	end
end

function AVRRangeWarningMesh:GenerateMesh()
	for i=1,self.numMarkers do
		self:GenerateMarkerMesh()
	end
	AVRMesh.GenerateMesh(self)
end

local markers={}
for i=1,40 do insert(markers,{}) end
function AVRRangeWarningMesh:OnUpdate(threed)
	local px,py
	local sx,sy=threed:GetUnitPosition("player")
	local dx,dy
	local numMarkers=0
	if GetNumRaidMembers()>0 then
		for i=1,40 do
			if not UnitIsUnit(raidID[i],"player") then 
				px,py=threed:GetUnitPosition(raidID[i])
				if px~=0 or py~=0 then
					dx,dy=sx-px,sy-py
					if dx*dx+dy*dy<=self.rangeSq then
						numMarkers=numMarkers+1
						markers[numMarkers].x,markers[numMarkers].y=px,py
					end
				end
			end
		end
	else
		for i=1,4 do
			px,py=threed:GetUnitPosition(partyID[i])
			if px~=0 or py~=0 then
				dx,dy=sx-px,sy-py
				if dx*dx+dy*dy<=self.rangeSq then
					numMarkers=numMarkers+1
					markers[numMarkers].x,markers[numMarkers].y=px,py
				end
			end
		end
	end
	if self.vertices==nil or numMarkers*(self.spokes+2)>#self.vertices then
		self.numMarkers=numMarkers
		self.vertices={}
		self.lines={}
		self.triangles={}
		self:GenerateMesh()
	end
	local spokeI,vertexI,v
	for i=1,numMarkers do
		spokeI=(i-1)*self.spokes+1
		vertexI=(i-1)*(self.spokes+2)+1
		for j=spokeI,spokeI+self.spokes-1 do
			self.triangles[j].a=nil
		end
		for j=vertexI,vertexI+self.spokes+1 do
			v=self.vertices[j]
			if v[9] then
				v[1]=v[1]-v[9]+markers[i].x
				v[2]=v[2]-v[10]+markers[i].y
			else
				v[1]=v[1]+markers[i].x
				v[2]=v[2]+markers[i].y
			end
			v[9],v[10]=markers[i].x,markers[i].y
		end
	end
	for i=numMarkers+1,self.numMarkers do
		spokeI=(i-1)*self.spokes+1
		vertexI=(i-1)*(self.spokes+1)+1
		for j=spokeI,spokeI+self.spokes-1 do
			self.triangles[j].a=0
		end
	end
	AVRMesh.OnUpdate(self,threed)
end

function AVRRangeWarningMesh:Pack()
	local s=AVRMesh.Pack(self)
	s.c="AVRRangeWarningMesh"
	s.ran=AVRPackDouble(self.range)
	s.rad=AVRPackDouble(self.radius)
	return s
end

function AVRRangeWarningMesh:Unpack(s)
	AVRMesh.Unpack(self,s)
	self.range=AVRUnpackDouble(s.ran) or 10
	self.rangeSq=self.range*self.range
	self.radius=AVRUnpackDouble(s.rad) or 1
	self.vertices=nil
end

AVRMeshClasses.AVRRangeWarningMesh=AVRRangeWarningMesh

function AVRRangeWarning(range,radius)
	local mesh=AVRRangeWarningMesh:New()
	mesh.name="Range warning "..mesh.range
	mesh:SetColor(nil,nil,nil,0.5)
	return mesh
end
